package tutorial;

import com.nativelibs4java.opencl.*;
import com.nativelibs4java.opencl.CLPlatform.DeviceFeature;
import com.nativelibs4java.util.*;
import java.io.File;
import java.io.IOException;
import java.nio.DoubleBuffer;
import org.apache.commons.math.complex.Complex;
import org.apache.commons.math.transform.FastFourierTransformer;

public class DFT {

    final CLQueue queue;
    final CLContext context;
    final CLProgram program;
    final CLKernel kernel;

    public DFT(CLQueue queue) throws IOException, CLBuildException {
        this.queue = queue;
        this.context = queue.getContext();

        String source = IOUtils.readText(new File("/Users/ochafik/Prog/Java/versionedSources/nativelibs4java/trunk/libraries/OpenCL/JavaCLTutorial/src/main/opencl/tutorial/DiscreteFourierTransformProgram.cl"));//DFT.class.getClassLoader().getResource("tutorial.DiscreteFourierTransformProgram.cl"));
        program = context.createProgram(source);
        kernel = program.createKernel("dft");
    }

    /**
     * Method that takes complex values in input (sequence of pairs of real and imaginary values) and 
     * returns the Discrete Fourier Transform of these values if forward == true or the inverse
     * transform if forward == false.
     */
    public synchronized DoubleBuffer dft(DoubleBuffer in, boolean forward) {
        assert in.capacity() % 2 == 0;
        int length = in.capacity() / 2;

        // Create an input CLBuffer that will be a copy of the NIO buffer :
        CLDoubleBuffer inBuf = context.createDoubleBuffer(CLMem.Usage.Input, in, true); // true = copy
        
        // Create an output CLBuffer :
        CLDoubleBuffer outBuf = context.createDoubleBuffer(CLMem.Usage.Output, length * 2);

        // Set the args of the kernel :
        kernel.setArgs(inBuf, outBuf, length, forward ? 1 : -1);
        
        // Ask for `length` parallel executions of the kernel in 1 dimension :
        CLEvent dftEvt = kernel.enqueueNDRange(queue, new int[]{ length });

        // Return an NIO buffer read from the output CLBuffer :
        return outBuf.read(queue, dftEvt);
    }

    /// Wrapper method that takes and returns double arrays
    public double[] dft(double[] complexValues, boolean forward) {
        DoubleBuffer outBuffer = dft(DoubleBuffer.wrap(complexValues), forward);
        double[] out = new double[complexValues.length];
        outBuffer.get(out);
        return out;
    }

    public static void main(String[] args) throws IOException, CLBuildException {
    		// Create a context with the best double numbers support possible :
    		// (try using DeviceFeature.GPU, DeviceFeature.CPU...)

        CLContext context = JavaCL.createBestContext();//DeviceFeature.DoubleSupport);
        
        // Create a command queue, if possible able to execute multiple jobs in parallel
        // (out-of-order queues will still respect the CLEvent chaining)
        CLQueue queue = context.createDefaultOutOfOrderQueueIfPossible();

        final DFT_float dft = new DFT_float(queue);
        //DFT2 dft = new DFT2(queue);
        final FastFourierTransformer fftApache = new FastFourierTransformer();

        // Create some fake test data :
        {
            int nWarm = 32;

            final double[] inDouble = createTestDoubleData(nWarm);
            final double[] inReal = toReal(inDouble);
            final float[] in = toFloat(inDouble);

            int warmups = 2500;
            System.out.println("Warming up...");
            for (int i = 0; i < warmups; i++) {
                testOpenCL(dft, in, i == 0);
                testApache(fftApache, inReal);
            }
        }

        System.gc();

        System.out.println("Testing...");
        int tests = 10;
        for (int nTest : new int[] { 1024 * 8, 1024 }) {
            final double[] inDouble = createTestDoubleData(nTest);
            final double[] inReal = toReal(inDouble);
            final float[] in = toFloat(inDouble);


            for (int i = 0; i < tests; i++) {
                time("Apache FFT (n = " + nTest + ")", new Runnable() { public void run() {
                    testApache(fftApache, inReal);
                }});
                System.gc();
            }
            for (int i = 0; i < tests; i++) {
                time("OpenCL DFT (n = " + nTest + ")", new Runnable() { public void run() {
                    testOpenCL(dft, in, false);
                }});
                System.gc();
            }
        }
    }
    static void time(String title, Runnable r) {
        long start = System.nanoTime();
        r.run();
        long time = System.nanoTime() - start;
        System.out.println("[" + title + "] = " + (time / 1000000.0) + " milliseconds");
    }
    static void testApache(FastFourierTransformer fftApache, double[] inReal) {
        // Transform the data (spatial -> frequency transform) :
        Complex[] transformedApache = fftApache.transform(inReal);

    }

    static void testOpenCL(DFT_float dft, float[] in, boolean check) {
        // Transform the data (spatial -> frequency transform) :

        dft.dft(in, true);

    }
    static void testOpenCL(DFT dft, double[] in, boolean check) {
        // Transform the data (spatial -> frequency transform) :
        
        double[] transformed = dft.dft(in, true);
        /*
        for (int i = 0; i < transformed.length / 2; i++) {
            // Print the transformed complex values (real + i * imaginary)
            System.out.println(transformed[i * 2] + "\t + \ti * " + transformed[i * 2 + 1]);
        }
         * 
         */

        if (check) {
            // Reverse-transform the transformed data (frequency -> spatial transform) :
            double[] backTransformed = dft.dft(transformed, false);

            // Check the transform + inverse transform give the original data back :
            double precision = 1e-5;
            for (int i = 0; i < in.length; i++) {
                if (Math.abs(in[i] - backTransformed[i]) > precision)
                    throw new RuntimeException("Different values in back-transformed array than in original array !");
            }
        }
    }
    static double[] toReal(double[] in) {
        double[] inReal = new double[in.length / 2];
        for (int i = 0; i < inReal.length; i++)
            inReal[i] = in[i / 2];

        return inReal;
    }
    static float[] toFloat(double[] in) {
        float[] out = new float[in.length];
        for (int i = 0; i < in.length; i++)
            out[i] = (float)in[i];

        return out;
    }
    static double[] createTestDoubleData(int n) {
        //int n = 1024 * 1024;
        double[] in = new double[2 * n];

        for (int i = 0; i < n; i++) {
            in[i * 2] = 1 / (double) (i + 1);
            in[i * 2 + 1] = 0;
        }
        return in;
    }
}
