// Enable double-precision floating point numbers support.
// Not all platforms / devices support this, so you may have to switch to floats.
#pragma OPENCL EXTENSION cl_khr_fp64 : enable

double2 rotate_double(double2 value, double c, double s) {
	return (double2)(
		dot(value, (double2)(c, -s)), 
		dot(value, (double2)(s, c))
	);
}

__kernel void dft(
	__global const double2 *in, // complex values input
	__global double2 *out,      // complex values output
	int offset,                 // offset in input buffer
	int stride,                 // stride in input buffer
	int inputLength,            // number of input values
	int outputLength,           // number of output values
	int sign)                   // sign modifier in the exponential :
	                            // 1 for forward transform, -1 for backward.
{
	// Get the varying parameter of the parallel execution :
	int i = get_global_id(0);
	
	// In case we're executed "too much", check bounds :
	if (i >= outputLength)
		return;
	
	// Initialize sum and inner arguments
	double2 tot = 0;
	double param = (-2 * sign * i) * M_PI / (double)inputLength;
	
	for (int k = 0; k < inputLength; k++) {
		double2 value = in[offset + k * stride];
		
		// Compute sin and cos in a single call : 
		double c;
		double s = sincos(k * param, &c);
		
		// This adds (value.x * c - value.y * s, value.x * s + value.y * c) to the sum :
		tot += rotate_double(value, c, s);
	}
	
	if (sign == 1) {
		// forward transform (space -> frequential)
		out[i] = tot;
	} else {
		// backward transform (frequential -> space)
		out[i] = tot / (double)inputLength;
	}
}

__kernel void fft_radix2(
	__global const double2 *in1, // complex values input
	__global const double2 *in2, // complex values input
	__global double2 *out,      // complex values output
	int length,                 // number of input and output values
	int sign)                   // sign modifier in the exponential :
	                            // 1 for forward transform, -1 for backward.
{
	// Get the varying parameter of the parallel execution :
	int i = get_global_id(0);
	
	// In case we're executed "too much", check bounds :
	if (i >= length)
		return;
	
	// Initialize sum and inner arguments
	double param = (-2 * sign) * M_PI / (double)length;
	
	// Compute sin and cos in a single call : 
	double c2;
	double s2 = sincos(param, &c2);
	
	out[i] = in1[i] + rotate_double(in2[i], c2, s2);
}

__kernel void fft_radix4(
	__global const double2 *in1, // complex values input
	__global const double2 *in2, // complex values input
	__global const double2 *in3, // complex values input
	__global const double2 *in4, // complex values input
	__global double2 *out,      // complex values output
	int length,                 // number of input and output values
	int sign)                   // sign modifier in the exponential :
	                            // 1 for forward transform, -1 for backward.
{
	// Get the varying parameter of the parallel execution :
	int i = get_global_id(0);
	
	// In case we're executed "too much", check bounds :
	if (i >= length)
		return;
	
	// Initialize sum and inner arguments
	double param = (-2 * sign) * M_PI / (double)length;
	
	// Compute sin and cos in a single call : 
	double c2, c3, c4;
	double s2 = sincos(param, &c2), s3 = sincos(param * 2, &c3), s4 = sincos(param * 3, &c4);
	
	out[i] = 
		in1[i] + 
		rotate_double(in2[i], c2, s2),
		rotate_double(in3[i], c3, s3),
		rotate_double(in4[i], c4, s4);
}
